/*------------------------------------------------------------------------------
| Part of Tweedledum Project.  This file is distributed under the MIT License.
| See accompanying file /LICENSE for details.
*-----------------------------------------------------------------------------*/
#pragma once

#include "Standard/H.h"
#include "Standard/Measure.h"
#include "Standard/P.h"
#include "Standard/Rx.h"
#include "Standard/Ry.h"
#include "Standard/Rz.h"
#include "Standard/S.h"
#include "Standard/Sx.h"
#include "Standard/Swap.h"
#include "Standard/T.h"
#include "Standard/U.h"
#include "Standard/X.h"
#include "Standard/Y.h"
#include "Standard/Z.h"
