/*------------------------------------------------------------------------------
| Part of Tweedledum Project.  This file is distributed under the MIT License.
| See accompanying file /LICENSE for details.
*-----------------------------------------------------------------------------*/
#pragma once

#include "a_star_swap_synth.h"
#include "all_linear_synth.h"
#include "cx_dihedral_synth.h"
#include "decomp_synth.h"
#include "diagonal_synth.h"
#include "gray_synth.h"
#include "lhrs_synth.h"
#include "linear_synth.h"
#include "pkrm_synth.h"
#include "pprm_synth.h"
#include "sat_linear_synth.h"
#include "sat_swap_synth.h"
#include "spectrum_synth.h"
#include "steiner_gauss_synth.h"
#include "transform_synth.h"
#include "xag_synth.h"
