/*------------------------------------------------------------------------------
| Part of Tweedledum Project.  This file is distributed under the MIT License.
| See accompanying file /LICENSE for details.
*-----------------------------------------------------------------------------*/
#pragma once

#include "Decomposition/barenco_decomp.h"
#include "Decomposition/bridge_decomp.h"
#include "Decomposition/one_qubit_decomp.h"
#include "Decomposition/parity_decomp.h"
