/*------------------------------------------------------------------------------
| Part of Tweedledum Project.  This file is distributed under the MIT License.
| See accompanying file /LICENSE for details.
*-----------------------------------------------------------------------------*/
#pragma once

#include "Utility/inverse.h"
#include "Utility/reverse.h"
#include "Utility/shallow_duplicate.h"
